/*   FILE: pam_keyring.c
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 4 June 2004
 * AUTHOR: Jonathan Nettleton <jon.nettleton@gmail.com>
 *   DATE: 13 March 2006
 *
 * Copyright (C) 2004 W. Michael Petullo <mike@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 */

#include <config.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <syslog.h>
#include <glib.h>
#include <pwd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>

#include <security/pam_modules.h>
#include <security/_pam_macros.h>
#include <security/pam_ext.h>
#include <security/pam_modutil.h>

#include <gnome-keyring-1/gnome-keyring.h>

#define PAM_DEBUG_ARG				0x0001
#define PAM_USE_FPASS_ARG			0x0002
#define PAM_TRY_FPASS_ARG			0x0004
#define PAM_USE_AUTHTOK 			0x0008

#define MISTYPED_PASS "Sorry, passwords do not match"
#define MAX_PASSWD_TRIES    3

#define _pam_delete(xx)     \
{               \
    _pam_overwrite(xx); \
    _pam_drop(xx);      \
}

/* this is a front-end for module-application conversations */

int _make_remark(pam_handle_t * pamh, unsigned int ctrl,
            int type, const char *text)
{
    int retval = PAM_SUCCESS;

    retval = pam_prompt(pamh, type, NULL, "%s", text);
    return retval;
}

/* ============================ preexec_t ================================== */
typedef struct preexec_t {
	const char *user;
	const char *evar;
} preexec_t;

/* ============================ free_cb () ================================= */
/* INPUT: pamh; data, must point to malloced memory of be NULL; errcode
 * SIDE AFFECTS: if data does not point to NULL then it is freed
 * NOTE: this is registered as a PAM callback function */
void free_cb(pam_handle_t * pamh, void *data, int errcode)
{
	if (data) {
		g_free(data);
	}
}

/* ============================ _pam_parse () ========================== */
/* INPUT: argc and argv, standard main()-type arguments
 * SIDE AFFECTS: global args are initialized, based on argc and argv */
static int
_pam_parse (const pam_handle_t *pamh, int argc,
            const char **argv, const char **keyring)
{
	int ctrl;

	*keyring = NULL;
	
	for (ctrl = 0; argc-- > 0; ++argv) {
		if (!strcmp(*argv,"debug"))
			ctrl |= PAM_DEBUG_ARG;
		else if (!strcmp(*argv,"use_first_pass"))
			ctrl |= PAM_USE_FPASS_ARG;
		else if (!strcmp(*argv,"try_first_pass"))
			ctrl |= PAM_TRY_FPASS_ARG;
		else if (!strcmp(*argv,"use_authtok"))
			ctrl |= PAM_USE_AUTHTOK;
		else if (!strncasecmp(*argv,"keyring=", 8))
			{
				*keyring = (*argv) + 8;
				if ( **keyring  == '\0' ) {
					*keyring = NULL;
					pam_syslog(pamh , LOG_ERR,
										"keyring= specification missing argument - using default");
				}
			}				
		 else
        	{
          		pam_syslog(pamh, LOG_ERR, "unknown option: %s", *argv);
        	}
	}		
	return ctrl;
}

/* ============================ read_password () =========================== */
/* INPUT: pamh; prompt1
 * SIDE AFFECTS: pass points to PAM's (user's) response to prompt
 * OUTPUT: any PAM error code encountered or PAM_SUCCESS
 * NOTE:   adapted from pam_userdb/pam_userdb.c
 */
/*
 * obtain a password from the user
 */

int _keyring_read_password(pam_handle_t * pamh
			,unsigned int ctrl
			,const char *comment
			,const char *prompt1
			,const char *prompt2
			,unsigned int authtok_flag 
			,const void **pass)
{
    
	int retval = PAM_SUCCESS;
	char *token;

	/*
	 * make sure nothing inappropriate gets returned
	 */

	*pass = token = NULL;

	/*
	 * should we obtain the password from a PAM item ?
	 */

	if ((ctrl & PAM_USE_FPASS_ARG) || (ctrl & PAM_TRY_FPASS_ARG)) 
    {
		retval = pam_get_item(pamh, authtok_flag, pass);
		if (retval != PAM_SUCCESS)  {
			/* very strange. */
			pam_syslog(pamh, LOG_ALERT, "pam_get_item returned error to keyring-read-password" );
			return retval;
		} else if (*pass != NULL) {
            /* we have a password! */
			return PAM_SUCCESS;
		} else if ((ctrl & PAM_USE_FPASS_ARG) == 1) {
            /* didn't work */ 
			return PAM_AUTHTOK_RECOVERY_ERR;	 
		}
	}
    
	/*
	 * getting here implies we will have to get the password from the
	 * user directly.
	 */

	int replies=1;
	char *resp[2] = { NULL, NULL };

	if (retval == PAM_SUCCESS) {
		retval = pam_prompt(pamh, PAM_PROMPT_ECHO_OFF,
		    &resp[0], "%s", prompt1);

		if (retval == PAM_SUCCESS && prompt2 != NULL) {
			retval = pam_prompt(pamh, PAM_PROMPT_ECHO_OFF,
			    &resp[1], "%s", prompt2);
			++replies;
		}
	}

	if (resp[0] != NULL && resp[replies-1] != NULL) {
		/* interpret the response */

		if (retval == PAM_SUCCESS) {	/* a good conversation */

			token = resp[0];
			if (token != NULL) {
				if (replies == 2) {
					/* verify that password entered correctly */
					if (strcmp(token, resp[replies - 1])) {
						/* mistyped */
						retval = PAM_AUTHTOK_RECOVERY_ERR;
						_make_remark(pamh, ctrl, PAM_ERROR_MSG, MISTYPED_PASS);
					}
				}
			} else {
				pam_syslog(pamh, LOG_NOTICE,
					    "could not recover authentication token");
			}
		}
	} else {
		retval = (retval == PAM_SUCCESS)
		    ? PAM_AUTHTOK_RECOVERY_ERR : retval;
	}

	resp[0] = NULL;
	if (replies > 1)
		_pam_delete(resp[1]);

	if (retval != PAM_SUCCESS) {
		_pam_delete(token);

		if (ctrl & PAM_DEBUG_ARG)
			pam_syslog(pamh, LOG_DEBUG, "unable to obtain a password");
		return retval;
	}
	/* 'token' is the entered password
	we store this password as an item */

	retval = pam_set_item(pamh, authtok_flag, token);
	_pam_delete(token);	/* clean it up */
    
	if (retval != PAM_SUCCESS ||
        (retval = pam_get_item(pamh, authtok_flag, pass))
	    != PAM_SUCCESS) {

		*pass = NULL;
		pam_syslog(pamh, LOG_CRIT, "error manipulating password");
		return retval;
	}
	return PAM_SUCCESS;
}


static int
obtain_authtok(pam_handle_t *pamh)
{
    char *resp;
    const void *item;
    int retval;

    retval = pam_prompt(pamh, PAM_PROMPT_ECHO_OFF, &resp, "Password: ");

    if (retval != PAM_SUCCESS)
        return retval;

    if (resp == NULL)
        return PAM_CONV_ERR;

    /* set the auth token */
    retval = pam_set_item(pamh, PAM_AUTHTOK, resp);

    /* clean it up */
    _pam_overwrite(resp);
    _pam_drop(resp);

    if ( (retval != PAM_SUCCESS) ||
         (retval = pam_get_item(pamh, PAM_AUTHTOK, &item))
         != PAM_SUCCESS ) {
        return retval;
    }

    return retval;
}


/* ============================ preexec () ================================= */
/* INPUT: user, a valid user name
 * SIDE AFFECTS: the calling process's uid, gid, euid and homedir are set to user's
 * NOTE: this is required because gnome-keyring-daemon restricts connections
 *       to its owner.
 */
static void preexec(const gpointer data)
{
	struct passwd *passwd_ent;

	assert(data);
	assert(((preexec_t *) data)->user);

	passwd_ent = getpwnam((char *) ((preexec_t *) data)->user);
	if (!passwd_ent) {
		syslog(LOG_ERR, "pam_keyring: error looking up user information for %s", ((preexec_t *) data)->user);
		exit(EXIT_FAILURE);
	}
	/* must exit because g_spawn_sync does not check return value before
	 * executing argument vector */
	if (setgid(passwd_ent->pw_gid) == -1) {
		syslog(LOG_ERR, "pam_keyring: error setting gid (%s)",  strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (setuid(passwd_ent->pw_uid) == -1) {
		syslog (LOG_ERR, "pam_keyring: error setting uid (%s)", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (seteuid(passwd_ent->pw_uid) == -1) {
		syslog (LOG_ERR, "pam_keyring: error setting euid: %s", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (setenv( "HOME", passwd_ent->pw_dir, 1) == -1) {
		syslog (LOG_ERR, "pam_keyring: error setting home: %s",  passwd_ent->pw_dir);
		exit(EXIT_FAILURE);
	}
	if (((preexec_t *) data)->evar)
		/* putenv does not want const char *, only char * */
		putenv((char *) ((preexec_t *) data)->evar);
}

/* ============================ keyring_daemon_stop () ===================== */
/* INPUT: data, a valid preexec_t name; pid, the pid of gnome-keyring-daemon
 * SIDE AFFECTS: a SIGTERM is sent to the gnome-keyring-daemon process
 * OUTPUT: TRUE if the SIGTERM is sent, otherwise FALSE
 */
static int keyring_daemon_stop(pam_handle_t *pamh, const preexec_t * data, pid_t pid)
{
	char *cmd_line;
	gchar **argv;
	GError *err = NULL;
	int retval = TRUE, status;

	assert(data != NULL);
	assert(data->user != NULL);

	cmd_line = g_strdup_printf("%s %d", KILL, pid);

	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: error parsing %s", cmd_line);
		goto _return;
	}
	if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, preexec,
			  (void *) data, NULL, NULL, &status, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: failed to run kill: %s", err->message);
		g_error_free(err);
		goto _return;
	}
	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: failed to execute code, exit code: %d",
							 WEXITSTATUS(status));
		retval = FALSE;
		goto _return;
	}
      _return:
	return retval;
}

/* ============================ keyring_daemon_start () ==================== */
/* INPUT: pamh, a valid pam_handle_t *; data, a valid user preexec_t
 * SIDE AFFECTS: gnome-keyring-daemon is started
 * OUTPUT: gnome-keyring-daemon's PID or 0 on error
 *         data->evar is set to g-k-d's i/o socket
 */
static pid_t keyring_daemon_start(pam_handle_t * pamh, preexec_t * data)
{
	/* FIXME: for some reason, gnome-keyring-daemon runs until pam_sm_authenticate
	   returns and then stops before the user's session starts when using with
	   login (not, ie., su).  Is the login process exiting and causing this?
	 */
	GError *err = NULL;
	char *standard_out = NULL;
	char **lines;
	int status;
	long pid;
	pid_t fnval = 0;
	char *pid_str, *end;
	gchar **argv = NULL;

	assert(pamh != NULL);
	assert(data != NULL);
	assert(data->user != NULL);
	if (!g_shell_parse_argv(GNOME_KEYRING_DAEMON, NULL, &argv, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: error parsing: %s", GNOME_KEYRING_DAEMON);
		goto _return;
	}
	if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, preexec,
			  (void *) data, &standard_out, NULL, &status, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: failed to run gome-keyring-daemon: %s", err->message);
		g_error_free(err);
		goto _return;
	}
	if (standard_out != NULL) {
		lines = g_strsplit(standard_out, "\n", 3);
		if (lines[0] != NULL &&
		    lines[1] != NULL &&
		    g_str_has_prefix(lines[1], "GNOME_KEYRING_PID=")) {
			pid_str = lines[1] + strlen("GNOME_KEYRING_PID=");
			pid = strtol(pid_str, &end, 10);
			if (end != pid_str) {
				fnval = pid;
				/* set variable for ulock operation */
				data->evar = g_strdup(lines[0]);
				/* FIXME: memory leak */
				/* ensure variable will be set in login session */
				if (pam_putenv (pamh, g_strdup(lines[0])) != PAM_SUCCESS) {
					fnval = 0;
					pam_syslog(pamh, LOG_ERR, "pam_keyring: error setting %s", lines[0]);
					goto _return;
				}
			}
		}
		g_strfreev(lines);
	} else {
		/* daemon failed for some reason */
		pam_syslog(pamh, LOG_ERR, "pam_keyring: gnome-keyring-daemon failed to start correctly, exit code: %d", WEXITSTATUS(status));
		pam_syslog(pamh, LOG_ERR, "pam_keyring: gnome-keyring-daemon returned this to STDOUT: %s", standard_out);
	}
	g_free(standard_out);
      _return:
	g_strfreev(argv);
	return fnval;
}

/* ============================ pam_keyring_unlock () ================================== */
/* INPUT: data, authtok, keyring
 * SIDE AFFECTS: user's GNOME keyring is unlocked
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
static int pam_keyring_unlock(pam_handle_t * pamh, const preexec_t * data,
		  const void *authtok, const char *keyring)
{
	gchar **argv = NULL;
	GError *err = NULL;
	int retval = PAM_SUCCESS;
	int standard_in = -1, standard_err = -1;
	char *cmd_line = NULL;
    GIOStatus result;
    gsize len;

	assert(data);
	assert(data->user);
	assert(authtok);

	if (keyring == NULL) {
        keyring = "default";
		cmd_line =
	    		g_strconcat(PAM_KEYRING_TOOL, " -u -s", NULL);
	} else {
		cmd_line =
	    		g_strconcat(PAM_KEYRING_TOOL, " -u  -s --keyring=", keyring, NULL);
	}
	
	pam_syslog(pamh, LOG_WARNING, "pam_keyring: going to execute %s", cmd_line);
	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: error parsing %s", cmd_line);
		goto _return;
	}
	if (g_spawn_async_with_pipes
	    (NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
	     preexec, (void *) data, NULL, &standard_in, NULL, &standard_err,
	     &err)) {
            GIOChannel *input;
            GIOChannel *error;
            gchar *error_buf = g_new(char,1); 

            error = g_io_channel_unix_new(standard_err);
            input = g_io_channel_unix_new(standard_in);
            g_io_channel_set_encoding (input, NULL, &err);
            g_io_channel_set_buffered (input, FALSE);
            
            result = g_io_channel_write_chars (input, authtok, -1, NULL, NULL);
            result = g_io_channel_write_chars (input, "\n", -1, NULL, NULL);
            
            if ( result == G_IO_STATUS_NORMAL) {
                result = g_io_channel_read_chars (error, error_buf, 1, &len, NULL);
            }
      
            g_io_channel_shutdown (error, FALSE, NULL);
            g_io_channel_unref (error);
    
            g_io_channel_shutdown (input, FALSE, NULL);
            g_io_channel_unref (input);
    } else {
            pam_syslog(pamh, LOG_WARNING, "pam_keyring: Error spawning pam-keyring-tool");
            retval = PAM_SERVICE_ERR;
            goto _return;
    }
    
    if (result == G_IO_STATUS_ERROR) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: the unlock sub-process didn't spawn correctly");
        goto _return;
    }
    
    if ( len ) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: PASSWORD failed to unlock the %s keyring", keyring);
        retval = PAM_SERVICE_ERR;
        goto _return;
    } else {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: PASSWORD successfully unlocked the %s keyring", keyring);
    }
            
      _return:
      
	g_strfreev(argv);
	g_free(cmd_line);
	return retval;
}

/* ============================ pam_verify_keyring_password () ================= */
/* INPUT: pamh, a valid pam_handle_t *; authtok, an authentication token
 * SIDE AFFECTS: the default GNOME keyring's password is set to authtok
 * OUTPUT: PAM_SUCCESS or PAM_SERVICE_ERR
 */
static int pam_verify_keyring_password(pam_handle_t * pamh, const preexec_t * data,
				const void *oldauthtok, const void *keyring)
{
	gchar **argv = NULL;
	GError *err = NULL;
	int retval = PAM_SUCCESS;
	int standard_in = -1, standard_err = -1;
	char *cmd_line = NULL;
    GIOStatus result;
    gsize len;
    
	assert(data);
	assert(data->user);
	assert(oldauthtok);

	if (keyring == NULL) {
        keyring = "default";
        cmd_line =
	    		g_strconcat(PAM_KEYRING_TOOL, " -t -s", NULL);
	} else {
		cmd_line =
	    		g_strconcat(PAM_KEYRING_TOOL, " -t  -s --keyring=", keyring, NULL);
	}
	
	pam_syslog(pamh, LOG_WARNING, "pam_keyring: going to execute %s", cmd_line);
	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: error parsing %s", cmd_line);
		goto _return;
	}
    
    if (g_spawn_async_with_pipes
        (NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
         preexec, (void *) data, NULL, &standard_in,  NULL, &standard_err,
         &err)) {
            GIOChannel *input;
            GIOChannel *error;
            gchar *error_buf = g_new(char, 1); 

            input = g_io_channel_unix_new(standard_in);
            error = g_io_channel_unix_new(standard_err);

            g_io_channel_set_encoding (input, NULL, &err);
            g_io_channel_set_buffered (input, FALSE);
            
            result = g_io_channel_write_chars (input, oldauthtok, -1, NULL, NULL);
            result = g_io_channel_write_chars (input, "\n", -1, NULL, NULL);
            
            if ( result == G_IO_STATUS_NORMAL) {
                result = g_io_channel_read_chars (error, error_buf, 1, &len, NULL);
            }            
    
            g_io_channel_shutdown (error, FALSE, NULL);
            g_io_channel_unref (error);
    
            g_io_channel_shutdown (input, FALSE, NULL);
            g_io_channel_unref (input);
    } else {
            pam_syslog(pamh, LOG_WARNING, "pam_keyring: Error spawning pam-keyring-tool");
            retval = PAM_SERVICE_ERR;
            goto _return;
    }
    
    if (result == G_IO_STATUS_ERROR) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: the check sub-process didn't spawn correctly");
        goto _return;
    }
    
    if ( len ) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: OLD_PASSWORD did not validate against the keyring");
        retval = PAM_SERVICE_ERR;
        goto _return;
    } else {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: OLD_PASSWORD validated against the keyring");
    }
    
    _return:
    
	g_strfreev(argv);
	g_free(cmd_line);
	return retval;
}

/* ============================ pam_change_keyring_password () ================= */
/* INPUT: pamh, a valid pam_handle_t *; authtok, an authentication token
 * SIDE AFFECTS: the default GNOME keyring's password is set to authtok
 * OUTPUT: PAM_SUCCESS or PAM_SERVICE_ERR
 */
static int pam_change_keyring_password(pam_handle_t * pamh, const preexec_t * data,
				const void *oldauthtok,  const void *authtok, const char *keyring)
{
	gchar **argv = NULL;
	GError *err = NULL;
	int retval = PAM_SUCCESS;
	int standard_in = -1, standard_err = -1;
	char *cmd_line = NULL;
    GIOStatus result;
    gsize len;

	assert(data);
	assert(data->user);
	assert(oldauthtok);
	assert(authtok);

	if (keyring == NULL) {
        keyring = "default";
		cmd_line =
	    		g_strconcat(PAM_KEYRING_TOOL, " -c -s", NULL);
	} else {
		cmd_line =
	    		g_strconcat(PAM_KEYRING_TOOL, " -c  -s --keyring=", keyring, NULL);
	}
	
	pam_syslog(pamh, LOG_WARNING, "pam_keyring: going to execute %s", cmd_line);
	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: error parsing %s", cmd_line);
		goto _return;
	}
	if (g_spawn_async_with_pipes
	    (NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
	     preexec, (void *) data, NULL, &standard_in, NULL, &standard_err,
	     &err)) {
            GIOChannel *input;
            GIOChannel *error;
            gchar *error_buf = g_new(char, 1); 

            error = g_io_channel_unix_new(standard_err);
            input = g_io_channel_unix_new(standard_in);
            g_io_channel_set_encoding (input, NULL, &err);
            g_io_channel_set_buffered (input, FALSE);
            
            result = g_io_channel_write_chars (input, oldauthtok, -1, NULL, NULL);
            result = g_io_channel_write_chars (input, "\n", -1, NULL, NULL);
            result = g_io_channel_write_chars (input, authtok, -1, NULL, NULL);
            result = g_io_channel_write_chars (input, "\n", -1, NULL, NULL);
            
            if ( result == G_IO_STATUS_NORMAL) {
                result = g_io_channel_read_chars (error, error_buf, 1, &len, NULL);
            } 
            g_io_channel_shutdown (error, FALSE, NULL);
            g_io_channel_unref (error);
    
            g_io_channel_shutdown (input, FALSE, NULL);
            g_io_channel_unref (input);  
     } else {     
            pam_syslog(pamh, LOG_WARNING, "pam_keyring: Error spawning pam-keyring-tool");
            retval = PAM_SERVICE_ERR;
            goto _return;
     }
     
     if (result == G_IO_STATUS_ERROR) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: the change sub-process didn't spawn correctly");
        goto _return;
    }
     
     if ( len ) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: change password for the %s keyring failed", keyring);
        retval = PAM_SERVICE_ERR;
        goto _return;
    } else {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: change password for the %s keyring succeeded", keyring);
    }
     
      _return:
	
    g_strfreev(argv);
	g_free(cmd_line);
	return retval;
}


/* ============================ pam_sm_authenticate () ===================== */
/* INPUT: this function is called by PAM
 * SIDE AFFECTS: user's GNOME keyring is unlocked
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 * NOTE:   adapted from pam_userdb/pam_userdb.c
 */
PAM_EXTERN int
pam_sm_authenticate(pam_handle_t * pamh, int flags,
		    int argc, const char **argv)
{
	const char *keyring = NULL;
	const void *password;
	pid_t *daemon_pid;
	int retval = PAM_AUTH_ERR, ctrl;
	preexec_t data = { NULL, NULL };

	ctrl = _pam_parse(pamh, argc, argv, &keyring);

	/* needed because gdm does not prompt for username as login does: */
	retval = pam_get_user(pamh, &data.user, NULL);
	if ((retval != PAM_SUCCESS) || (!data.user)) {
		pam_syslog(pamh, LOG_ERR, "can not get the username");
		retval = PAM_SERVICE_ERR;
		goto _return;
	}
	
	if ((ctrl & PAM_USE_FPASS_ARG) == 0 && (ctrl & PAM_TRY_FPASS_ARG) == 0) {
        /* Converse to obtain a password */
        retval = obtain_authtok(pamh);
        if (retval != PAM_SUCCESS) {
            pam_syslog(pamh, LOG_ERR, "can not obtain password from user");
            goto _return;
        }
	}
	
	/* Check if we got a password */
	retval = pam_get_item(pamh, PAM_AUTHTOK, &password);
	if (retval != PAM_SUCCESS || password == NULL) {
		if ((ctrl & PAM_TRY_FPASS_ARG) != 0) {
            retval = obtain_authtok(pamh);
            if (retval != PAM_SUCCESS) {
                pam_syslog(pamh, LOG_ERR, "can not obtain password from user");
                return retval;
            }
            retval = pam_get_item(pamh, PAM_AUTHTOK, &password);
        }
        if (retval != PAM_SUCCESS || password == NULL) {
            pam_syslog(pamh, LOG_ERR, "can not recover user password");
            return PAM_AUTHTOK_RECOVER_ERR;
        }
     }
	
	if (ctrl & PAM_DEBUG_ARG)
		pam_syslog(pamh, LOG_INFO, "Verify user `%s' with a password",
							data.user);
	
	if ((data.evar = getenv("GNOME_KEYRING_SOCKET")) != NULL) {
		pam_syslog(pamh, LOG_WARNING, "pam_keyring: daemon already exists (%s)", data.evar);
		/* FIXME: g_strconcat = memory leak? */
		/* ensure variable will be passed to next session */
		if (pam_putenv(pamh, g_strconcat("GNOME_KEYRING_SOCKET=",
						 data.evar,
						 NULL)) != PAM_SUCCESS) {
            pam_syslog(pamh, LOG_ERR, "pam_keyring: error setting GNOME_KEYRING_SOCKET");
			retval = PAM_SERVICE_ERR;
			goto _return;
		}
		retval = PAM_SUCCESS;
		goto _return;
	}

	pam_syslog(pamh, LOG_WARNING, "pam_keyring: starting gnome-keyring-daemon");
	daemon_pid = g_new0(pid_t, 1);	// FIXME: memory leak?
	if ((*daemon_pid = keyring_daemon_start(pamh, &data)) == 0) {
		retval = PAM_SERVICE_ERR;
		goto _return;
	}
	if ((retval =
	     pam_set_data(pamh, "pam_keyring_gkd_pid", daemon_pid,
			  free_cb)) != PAM_SUCCESS) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: %s\n", "error trying to save gnome-keyring-daemon PID");
		goto _return;
	}
	pam_syslog(pamh, LOG_WARNING, "pam_keyring: unlocking keyring");
	retval = pam_keyring_unlock(pamh, &data, password, keyring);
      _return:
	return retval;
}

/* ============================ pam_sm_chauthtok () ======================== */
/* INPUT: this function is called by PAM, gnome-keyring-daemon must be running
 * SIDE AFFECTS: user's GNOME keyring has its password changed
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t * pamh, int flags, int argc,
         const char **argv)
{
    const char *keyring = NULL;
    const char *announce = NULL;
    const void *password_old, *password_new;
    
    int retval = PAM_AUTH_ERR, ctrl;
    preexec_t data = { NULL, NULL };

    ctrl = _pam_parse(pamh, argc, argv, &keyring);
    
     if ((retval = pam_get_user(pamh, &data.user, NULL)) != PAM_SUCCESS) {
        pam_syslog(pamh, LOG_ERR, "pam_keyring: %s", "could not get user");
        /* do NOT return PAM_SERVICE_ERR or root will not be able 
         * to su to other users */
        goto _return;
    }
    
    if (!strcmp(data.user, "root")) {
        pam_syslog(pamh, LOG_WARNING, "pam_keyring: do nothing for root");
        retval = PAM_SUCCESS;
        goto _return;
    }
    
    if (flags & PAM_PRELIM_CHECK) {
        if (gnome_keyring_is_available()) {
            announce = g_strdup_printf("Changing password for %s.", data.user);       
            retval = _keyring_read_password(pamh, ctrl
                                                                    ,announce
                                                                    ,"(current) keyring password: "
                                                                    ,NULL
                                                                    ,PAM_OLDAUTHTOK
                                                                    ,&password_old);
            
            if (retval != PAM_SUCCESS) {
                pam_syslog(pamh, LOG_NOTICE, "password - (old) token not obtained");
                return retval;
            }
        } else {
            retval = PAM_TRY_AGAIN;
        }
        retval = pam_verify_keyring_password(pamh, &data, password_old, keyring);
    } else if ((flags & PAM_UPDATE_AUTHTOK) && gnome_keyring_is_available()) {
        
        
        retval = pam_get_item(pamh, PAM_OLDAUTHTOK, &password_old);
        
        if (retval != PAM_SUCCESS) {
            pam_syslog(pamh, LOG_NOTICE, "user not authenticated");
            return retval;
        }
        
        pam_syslog(pamh, LOG_NOTICE, "we are going to try and get a new password");

        if (ctrl & PAM_USE_AUTHTOK)
           ctrl |= PAM_USE_FPASS_ARG; 
        
        int retry = 0;
        retval = PAM_AUTHTOK_ERR;
        
        while ((retval != PAM_SUCCESS) && (retry++ < MAX_PASSWD_TRIES)) {
            /*
             * use_authtok is to force the use of a previously entered
             * password -- needed for pluggable password strength checking
             */

            retval = _keyring_read_password(pamh, ctrl, NULL
                                                                ,"Enter new keyring password: "
                                                                ,"Retype new keyring password: "
                                                                ,PAM_AUTHTOK
                                                                ,&password_new);

            if (retval != PAM_SUCCESS) {
                if (ctrl & PAM_DEBUG_ARG)
                    pam_syslog(pamh, LOG_ALERT, "password - new keyring password not obtained");
                    
                password_old = NULL;    /* tidy up */
                return retval;
            }

            /*
             * At this point we know who the user is and what they
             * propose as their new password. Verify that the new
             * password is acceptable.
             */

            if (*(const char *)password_new == '\0') {  /* "\0" password = NULL */
                password_new = NULL;
            }
        }
        
        /* get ready to start checking things against the keyring */
        preexec(&data);
        
        if (password_old) {
            retval = pam_verify_keyring_password(pamh, &data, password_old, keyring);
            if (retval != PAM_SUCCESS) {
                pam_syslog(pamh, LOG_NOTICE, "keyring password changed by another process");
                return retval;
            }
        }
        
        if ((password_new == NULL || password_old  == NULL ) 
                || !strcmp(password_old, password_new)) {
            if (ctrl & PAM_DEBUG_ARG)
                pam_syslog(pamh, LOG_INFO, "bad authentication token");
            _make_remark(pamh, ctrl, PAM_ERROR_MSG, password_new == NULL ?
                                    "No password supplied" :"Password unchanged");
            return PAM_AUTHTOK_ERR;
        }
        
        retval = pam_change_keyring_password(pamh, &data, password_old, password_new, keyring);
        
        password_old = password_new = NULL;
    } else {
        pam_syslog(pamh, LOG_ALERT,
                               "password received unknown request");
        retval = PAM_ABORT;
    }
    
    _return:
    return retval;
}

/* ============================ pam_sm_open_session () ===================== */
/* NOTE: placeholder function so PAM does not get mad */
PAM_EXTERN int
pam_sm_open_session(pam_handle_t * pamh, int flags,
		    int argc, const char **argv)
{
	return PAM_SUCCESS;
}

/* ============================ pam_sm_close_session () ==================== */
/* INPUT: this function is called by PAM
 * SIDE AFFECTS: user's gnome-keyring-daemon is sent a SIGTERM
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
PAM_EXTERN int
pam_sm_close_session(pam_handle_t * pamh, int flags, int argc,
		     const char **argv)
{
	preexec_t data = { NULL, NULL };
	const void *daemon_pid;
	int retval = PAM_SUCCESS;

	assert(pamh);

	pam_syslog(pamh, LOG_WARNING, "pam_keyring: received order to close session");

	if ((retval = pam_get_user(pamh, &data.user, NULL)) != PAM_SUCCESS) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: %s", "could not get user");
		/* do NOT return PAM_SERVICE_ERR or root will not be able 
		 * to su to other users */
		goto _return;
	}
	if (!strcmp(data.user, "root")) {
		pam_syslog(pamh, LOG_WARNING, "pam_keyring: do nothing for root");
		retval = PAM_SUCCESS;
		goto _return;
	}
	if ((retval =
	     pam_get_data(pamh, "pam_keyring_gkd_pid",
			  &daemon_pid)) != PAM_SUCCESS) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: %s", "could not retrieve gnome-keyring-deamon PID");
		retval = PAM_SUCCESS;
		goto _return;
	}
	if (keyring_daemon_stop(pamh, &data, *(int *) daemon_pid) == FALSE) {
		pam_syslog(pamh, LOG_ERR, "pam_keyring: error trying to kill gnome-keyring-daemon (%d)", *(int *) daemon_pid);
		retval = PAM_SERVICE_ERR;
		goto _return;
	}
      _return:
	return retval;
}

/* ============================ pam_sm_setcred () ========================== */
/* NOTE: placeholder function so PAM does not get mad */
PAM_EXTERN int
pam_sm_setcred(pam_handle_t * pamh, int flags, int argc, const char **argv)
{
	return PAM_SUCCESS;
}
